﻿#pragma TextEncoding = "UTF-8"
#pragma rtGlobals=3
#pragma version=2.00
#pragma IndependentModule=WiREloader
#pragma IgorVersion=8
#include <Resize Controls>

// updater headers
static constant kProjectID=21301 // the project node on IgorExchange
static strconstant ksShortTitle="WireLoader" // the project short title on IgorExchange

static strconstant ksPackageName = "WireLoader"
static strconstant ksPrefsFileName = "acwWireLoader.bin"
static constant kPrefsVersion = 100

// A loader for Renishaw Raman files of type wdf.
// https://www.wavemetrics.com/user/tony

// To adjust settings: Data > Load Waves > Packages > Wire Loader Settings...

// Map data are by default loaded as a 3D wave. 3D waves are compatible 
// with the hyperspectral browser project: 
// https:www.wavemetrics.com/node/22408 
// WiRE loader adds a wavenote to specify associated zwave and 
// photographic image. Wavenote is a keylist with keys Zwave and ImgName,
// key separator is : and list separator is \r 
// Zwave:wavename
// ImgWave:wavename
// These keys are compatible with the hyperspectral browser project.

// Alternatively, map files (including line scans) can be loaded as 
// multiple 1D spectra with a common wave containing wavenumbers. 
// Coordinates are recorded in wave notes. The map browser can then be 
// used to select spectra when an image has been saved in the WiRE file

// 1.22 adjustment to wave note format
// 1.20 delete package folder when closing panel
// 1.10 4/6/20 Reading of map coordinates now more reliable
//      Z positions are recorded for depth scans.
// 1.00 15/7/18 

// commandline options
// "WiREloader#Settings(Load3D=0|1)"
//	"WiREloader#Settings(PlotMap=0|1)"
//	"WiREloader#Settings(Verbose=0|1|2) // 2 = test mode"
//	"WiREloader#Settings(Overwrite=0|1|2) // 2 = ask"
//	"WiREloader#Settings(SetDefault=1)"

menu "Load Waves"
	"Load WiRE Files...", /Q, loadWiRE()
	submenu "Packages"
		"WiRE loader settings...", /Q, Settings()
	end
end

menu "Data"
	"WiRE map browser", /Q, makeMapBrowser()
end

structure PackagePrefs
	uint16 version // 2 bytes
	uint16 options // 2 bytes, 1 = load 3D waves, 2 = plot maps
	uint16 loadoptions // 2 bytes
	char verbose
	char overwrite // 1 byte, 0 = rename, 1 = yes, 2 = ask
	char suffix[10] // suffix for wavenumber wave
	char reserved[128 - 18]
endstructure

// set prefs structure to default values
function PrefsSetDefaults(STRUCT PackagePrefs &prefs)
	prefs.version     = kPrefsVersion
	prefs.options     = 3
	prefs.loadoptions = 0
	prefs.verbose     = 1
	prefs.overwrite   = 2
	prefs.suffix      = "_x"
	int i
	for(i=0;i<(128-18);i+=1)
		prefs.reserved[i] = 0
	endfor
end

function LoadPrefs(STRUCT PackagePrefs &prefs)
	LoadPackagePreferences/MIS=1 ksPackageName, ksPrefsFileName, 0, prefs
	if (V_flag!=0 || V_bytesRead==0 || prefs.version!=kPrefsVersion)
		PrefsSetDefaults(prefs)
		SavePackagePreferences ksPackageName, ksPrefsFileName, 0, prefs
	endif
end

function Settings([int Load3D, int PlotMap, int Verbose, int Overwrite, int SetDefault])
	
	STRUCT PackagePrefs prefs
	LoadPrefs(prefs)	
	int cmd = 0
	
	if (paramisdefault(SetDefault) == 0)
		PrefsSetDefaults(prefs)
		cmd = 1
	endif
	
	if (paramisdefault(Load3D) == 0)
		Load3D = Load3D != 0
		prefs.options = (prefs.options & ~1) | Load3D
		cmd = 1	
	endif
	
	if (paramisdefault(PlotMap) == 0)
		prefs.options = (prefs.options & ~1) | 2*(plotMap!=0)
		cmd = 1	
	endif
	
	if (paramisdefault(Verbose) == 0)
		prefs.verbose = Verbose
		cmd = 1
	endif
	
	if (paramisdefault(Overwrite) == 0)
		Overwrite = limit(Overwrite, 0, 2)
		prefs.overwrite = Overwrite
		cmd = 1
	endif
	
	if (cmd == 0)
		MakePrefsPanel()
		return 0
	endif

//	commandline options
//		print "WiREloader#Settings(Load3D=0|1)"
//		print "WiREloader#Settings(PlotMap=0|1)"
//		print "WiREloader#Settings(Verbose=0|1|2) // 2 = test mode"
//		print "WiREloader#Settings(Overwrite=0|1|2) // 2 = ask"
//		print "WiREloader#Settings(SetDefault=1)"
	
	SavePackagePreferences ksPackageName, ksPrefsFileName, 0, prefs
end

function MakePrefsPanel()
			
	STRUCT PackagePrefs prefs
	LoadPrefs(prefs)

	
	KillWindow/Z WirePrefsPanel
	
	STRUCT point screenOrigin
	GetScreenOrigin(screenOrigin)
		
	int WL = screenOrigin.h + 300
	int WT = screenOrigin.v + 200
	int width = 250
	int height = 170

	NewPanel/K=1/N=WirePrefsPanel/W=(WL,WT,WL+width,WT+height) as "WiRE Loader [version " + num2str(GetThisVersion()/100) +"]"
	ModifyPanel/W=WirePrefsPanel, fixedSize=1, noEdit=1
	
	variable x0 = 10
	variable y0 = 10
	
	variable font = 12

	CheckBox chk3D, pos={x0, y0}, fSize=font, title="Load maps as 3D waves"
	CheckBox chk3D, help={""}
	y0 += 20
	CheckBox chkPlotMap, pos={x0, y0}, fSize=font, title="Plot map data"
	CheckBox chkPlotMap, help={""}
	y0 += 20
	PopupMenu popOverwrite, pos={x0, y0}, fSize=font, title="Overwrite", value="Never;Without asking;Always ask"
	PopupMenu popOverwrite, help={""}
	
//	CheckBox chkOverwrite, pos={x0, y0}, fSize=font, title="Overwrite Without Warning"
//	CheckBox chkOverwrite, help={"Overwrite waves without warning"}
	y0 += 20
	CheckBox chkVerbose, pos={x0, y0}, fSize=font, title="Print to history"
	CheckBox chkVerbose, help={"Record file loading details in experiment history"}
	y0 += 20
	SetVariable svSuff, pos={x0,y0}, size={165,16}, title="Wavenumber Suffix", Proc=wireloader#Setvars
	SetVariable svSuff, help={"Suffix is appended to output wavenumber wave names"}, fSize=font
	y0 += 50
		
//	CheckBox chkSubfolders, pos={x0,y0}, title="Save Waves in Subfolders", mode=0, fSize=font
//	CheckBox chkSubfolders, help={"Save output waves in separate subfolder for each file"}

	Button btnDefaults, pos={x0,y0}, size={60,22}, title="Defaults", Proc=wireloader#buttons
	Button btnDefaults, help={"Set all controls to default values"}
	Button btnCancel, pos={120,y0}, size={55,22}, title="Cancel", Proc=wireloader#buttons
	Button btnSave, pos={190,y0}, size={50,22}, title="Save", Proc=wireloader#buttons
	
	SetControlValuesFromPrefs(prefs)
	
	PauseForUser WirePrefsPanel
	
	return 0
end

function SetControlValuesFromPrefs(STRUCT PackagePrefs &prefs)
	CheckBox chk3D win=WirePrefsPanel, value=prefs.options&1
	CheckBox chkPlotMap win=WirePrefsPanel, value=prefs.options&2
//	CheckBox chkOverwrite win=WirePrefsPanel, value=prefs.overwrite==1
	PopupMenu popOverwrite win=WirePrefsPanel, mode=prefs.overwrite + 1
	CheckBox chkVerbose win=WirePrefsPanel, value=prefs.verbose==1	
	SetVariable svSuff win=WirePrefsPanel, value=_STR:prefs.suffix
end

function GetPrefsControlValues(STRUCT PackagePrefs &prefs)
	controlinfo/W=WirePrefsPanel chk3D
	prefs.options = (prefs.options & ~1) | v_value
	controlinfo/W=WirePrefsPanel chkPlotMap
	prefs.options = (prefs.options & ~2) | (2 * v_value)
	controlinfo/W=WirePrefsPanel popOverwrite
	prefs.overwrite = v_value - 1
	controlinfo/W=WirePrefsPanel chkVerbose
	prefs.verbose = v_value
	controlinfo/W=WirePrefsPanel svSuff
	prefs.suffix = S_Value
end

function Buttons(STRUCT WMButtonAction &s) : ButtonControl

	if (s.eventcode != 2)
		return 0
	endif
	
	STRUCT PackagePrefs prefs
	strswitch(s.ctrlname)
		case "btnDefaults" :
			LoadPrefs(prefs)
			PrefsSetDefaults(prefs)
			SetControlValuesFromPrefs(prefs)
			break
		case "btnSave" :
			LoadPrefs(prefs)
			GetPrefsControlValues(prefs)
			SavePackagePreferences ksPackageName, ksPrefsFileName, 0, prefs
		case "btnCancel" :		
			killWindow $s.win
			break			
	endswitch

	return 0
end

function Setvars(STRUCT WMSetVariableAction &s) : SetVariableControl
	if (s.eventcode==2 && cmpstr(s.ctrlname, "svSuff")==0)
		if (!strlen(s.sval))
			SetVariable $s.ctrlname win=$s.win, value=_STR:"_x"
		elseif (strlen(s.sval) > 10)
			SetVariable $s.ctrlname win=$s.win, value=_STR:(s.sval)[0,9]
		endif
	endif	
	return 0
end

function loadWiRE()
	STRUCT PackagePrefs prefs
	LoadPrefs(prefs)
		
	variable i
	Open/D/R/MULT=1/F="WiRE Files (*.wdf):.wdf;"/M="Looking for WiRE files..." refnum
	wave/T w_selection = ListToTextWave(S_fileName, "\r")
	for (i=0;i<numpnts(w_selection);i+=1)
		loadWDF(w_selection[i], prefs)
	endfor
	return 1
end

function loadWDF(string strPath, STRUCT PackagePrefs &prefs)
			
	string strBlockName = "", strData = "", strFileType = "", cmd = ""
	variable refnum = 0, BlockLength = 0, blockStart = 0, i = 0
	variable isMap = 0, haveImageCoords = 0, haveMapCoords = 0, haveORGNcoords = 0
	variable ORGNstart = 0, ORGNend = 0, WXDMstart = 0, WXDMend = 0, WMAPstart = 0
		
	Make/free/N=0/D w_x, w_y, w_z // double precision waves to hold coordinates
	Make/free/N=0 w_wavenumber, w_data // single precision waves to hold data from XLST and DATA blocks
	Make/free/N=6 MapInfo // data from WMAP block
	Make/free/B/N=0 w_jpeg // wave to hold any jpeg binary embedded in wdf
	Make/free/N=4/D ImageCoords
	wave/Z img = $""
	
	Open/R/Z refnum as strPath
	FStatus refnum // sets V_logEOF, S_fileName
	
	if (prefs.verbose && v_flag==0)
		printf "WiRE loader: couldn't open %s\r", strPath
		return 0
	elseif (prefs.verbose)
		printf "* WiRE loader reading from %s%s *\r", S_path, S_fileName
	endif
	
	string strNote = ""
	string strBaseName = RemoveEnding(s_fileName, ".wdf")
	strBaseName = CleanupName(strBaseName, 0)
	strBaseName = RemoveEnding(strBaseName, "_")
	
	int load3D = prefs.options & 1
	int PlotMap = prefs.options & 2
	
	do // loop through blocks in file
		BlockLength = 0
		strBlockName = ""
		strBlockName = PadString(strBlockName, 4, 0)
		
		FBinRead refnum, strBlockName
		if (i==0 && cmpstr(strBlockName, "WDF1"))
			printf "WiRE loader: %s does not appear to be a wdf file\r", s_fileName
			Close refnum
			return 0
		endif
		FSetPos refnum, blockStart+8
		FBinRead/B=3/F=3/U refnum, BlockLength // 4 bytes
		
		if (BlockLength == 0)
			break
		endif
		
		strswitch(strBlockName)
			case "WDF1":
				FSetPos refnum, blockStart+240
				strFileType = PadString(strFileType, 6, 0)
				FBinRead refnum, strFileType
				break
			case "XLST":
				Redimension/S/N=((BlockLength-24)/4) w_wavenumber
				FSetPos refnum, blockStart+24
				FBinRead/B=3 refnum, w_wavenumber
				break
			case "DATA":
				// if file contains map data this will be a concatenation of spectra
				Redimension/S/N=((BlockLength-16)/4) w_data
				FSetPos refnum, blockStart+16
				FBinRead/B=3 refnum, w_data
				break
			case "WMAP":
				isMap = 1
				WMAPstart = blockStart
				FSetPos refnum, blockStart+24
				FBinRead/B=3/F=4 refnum, MapInfo // 6 points, mapX0, mapY0, unknown1, mapXstep, mapYstep, unknown2, 32 bit floats in wdf
				Make/free/N=2 w_temp
				FBinRead/B=3/F=3/U refnum, w_temp
				Concatenate/NP {w_temp}, MapInfo // 7th and 8th points are mapWidth, mapHeight, 32 bit integers in wdf
				break
			case "ORGN":
				ORGNstart = blockStart
				ORGNend = blockstart+blockLength
				break
			case "WHTL":
				Redimension/N=((BlockLength-16)) w_jpeg
				FSetPos refnum, blockStart+16
				FBinRead/B=3 refnum, w_jpeg	 // read jpeg binary into w_jpeg
				break
			case "WXDM": // here is where we find image registration
				WXDMstart = blockstart
				WXDMend = blockstart+blockLength
				break
		endswitch
		
		if (prefs.verbose > 1)
			printf "%s start: %.10g length: %.10g bytes\r", strBlockName, blockStart, blockLength
		endif
		
		blockStart = blockStart + BlockLength
		
		if (blockStart >= V_logEOF)
			break
		endif
		
		FSetPos refnum, blockStart
		i += 1
	while (i < 1000)
	
	if (prefs.verbose && cmpstr(strFileType, "Depth ")==0)
		Print "File appears to be depth series acquisition"
	elseif (prefs.verbose && cmpstr(strFileType, "Single")==0)
		Print "File appears to be single scan measurement"
	endif
	
	variable Xlength = numpnts(w_wavenumber)
	variable dataLength = numpnts(w_data)
	variable numSpectra = dataLength/Xlength
	if (mod(datalength, xlength) != 0)
		Print "WiRE loader errror: mismatched wavelengths in " + S_fileName
		numSpectra = 0
	endif
		
	if (ORGNstart && numSpectra)
		haveORGNcoords = readXYZpositions(w_x, w_y, w_z, numSpectra, ORGNstart, ORGNend, refnum, prefs)
	endif
	
	if (isMap && ORGNstart)
		haveMapCoords = (numpnts(w_x)==numSpectra && numpnts(w_y)==numSpectra)
	endif
	
	if (numpnts(w_jpeg) && WXDMstart)
		if (haveMapCoords)
			Make/free/N=4 w_MapLimits={WaveMin(w_x),WaveMax(w_x),WaveMin(w_y),WaveMax(w_y)}
			haveImageCoords = readImageCoords(ImageCoords, WXDMstart, WXDMend, refnum, prefs, MapLimits=w_MapLimits)
		else
			haveImageCoords = readImageCoords(ImageCoords, WXDMstart, WXDMend, refnum, prefs)
		endif
	endif

	Close refnum // done with reading data from file
		
	if (numpnts(w_data) == 0)
		printf "WiRE loader error: could not locate data block in %s\r", S_fileName
		return 0
	endif
	
	if (numpnts(w_wavenumber) == 0)
		printf "WiRE loader error: could not locate wavenumber block in %s\r", S_fileName
		return 0
	endif
	
	if (numspectra == 0)
		return 0
	endif
		
	string xName = strBaseName + prefs.suffix
	
	// check for waves to be overwritten
	string overwriteList = SelectString(exists(xName), "", xName+";")

	if ((numpnts(w_jpeg)>0))
		overwriteList += SelectString(exists(strBaseName+"_img"), "", strBaseName+"_img;")
	endif
	
	overwriteList += SelectString(numpnts(w_x)>1 && exists(strBaseName+"_mapX"), "", strBaseName+"_mapX;")
	overwriteList += SelectString(numpnts(w_y)>1 && exists(strBaseName+"_mapY"), "", strBaseName+"_mapY;")
	overwriteList += SelectString(numpnts(w_z)>1 && exists(strBaseName+"_mapZ"), "", strBaseName+"_mapZ;")
	
	if (!load3D && numSpectra>1)
		for (i=0; i<numSpectra; i+=1)
			overwriteList += SelectString(exists(strBaseName+"_"+num2str(i)), "", strBaseName+"_"+num2str(i)+";")
		endfor
	else
		overwriteList += SelectString(exists(strBaseName), "", strBaseName+";")
	endif
	
	if (ItemsInList(overwriteList))
		if (prefs.overwrite == 0)
			cmd = s_fileName + ": waves already exist. Current settings do not allow overwriting."
			if (prefs.verbose)
				Print cmd 
			endif
			cmd += "\r\rData > Load Waves > Packages > Wire Loader Settings..."
			DoAlert 0, cmd
			return 0
		elseif (prefs.overwrite == 2)
			DoAlert 1, DoAlertString(OverwriteList)
			if (prefs.verbose && v_flag==2)
				if (prefs.verbose)
					Print s_fileName + " load aborted"
				endif
				return 0
			endif
		endif
	endif
	
	if (prefs.verbose && isMap)
		printf "Map file with %g spectra, ", numSpectra
		printf "origin = (%g,%g), step = (%g,%g), size = (%d,%d)\r", MapInfo[0], MapInfo[1], MapInfo[3], MapInfo[4], MapInfo[6], MapInfo[7]
	endif
	
	if (prefs.verbose > 1)
		printf "Single precision Raman shift data from XLST block, %d points\r", numpnts(w_wavenumber)
		printf "Single precision intensity data from DATA block, %d points\r", numpnts(w_data)
	endif
	
	// save Raman shift wave
	Duplicate/O w_wavenumber $xName/wave=wx
	if (prefs.verbose)
		printf "Raman shifts saved as %s\r", xName
	endif
	
	if (numSpectra == 1) // for single point save intensity wave and we're done
		if (haveORGNcoords)
			sprintf cmd "Coordinates:"
			if (numpnts(w_x) == 1)
				sprintf cmd "%sX=%f;", cmd, w_x[0]
			endif
			if (numpnts(w_y) == 1)
				sprintf cmd "%sY=%f;", cmd, w_y[0]
			endif
			if (numpnts(w_z) == 1)
				sprintf cmd "%sZ=%f;", cmd, w_z[0]
			endif
			cmd = RemoveEnding(cmd, ";")+"\r"
		endif
		sprintf cmd, "%sXwave:%s", cmd, xName
		note w_data cmd
		Duplicate/O w_data $strBaseName
		if (prefs.verbose)
			printf "Intensity spectrum saved as %s\r", strBaseName
		endif
		return 1
	endif
	
	if (isMap && load3D)
		make/O/N=(MapInfo[6], MapInfo[7], numpnts(w_data)/numSpectra) $strBaseName /wave=w3D
		w3D = w_data[(p+q*MapInfo[6])*xlength + r]
		SetScale/P x, MapInfo[0], mapInfo[3], w3D
		SetScale/P y, MapInfo[1], mapInfo[4], w3D
		strNote = note(w3D)
		strNote = ReplaceStringByKey("Zwave", strNote, xName, ":", "\r")
		note/K w3D, strNote
		if (prefs.verbose)
			printf "Intensity data saved as %s\r", strBaseName
		endif
	else	
		i=0
		do
			Make/O/N=(xlength) $(strBaseName+"_"+num2str(i)) /wave=ithDataWave
			ithDataWave = w_data[p+i*xlength]
			cmd = ""
			if (haveORGNcoords)
				sprintf cmd "Coordinates:"
				if (numpnts(w_x)>i)
					sprintf cmd "%sX=%f;", cmd, w_x[i]
				endif
				if (numpnts(w_y)>i)
					sprintf cmd "%sY=%f;", cmd, w_y[i]
				endif
				if (numpnts(w_z)>i)
					sprintf cmd "%sZ=%f;", cmd, w_z[i]
				endif
				cmd = RemoveEnding(cmd, ";")+"\r"
			endif
			sprintf cmd, "%sXwave:%s", cmd, xName
			note ithDataWave cmd
			i += 1
		while ((i*xlength) < datalength)
		if (prefs.verbose)
			printf "%d intensity spectra saved as %s_N where N increments from 0 to %d\r", i, strBaseName, i-1
		endif
	endif
	
	// save image
	if (numpnts(w_jpeg))
		wave/Z img = saveImage(w_jpeg, strBaseName, prefs)
	endif
	
	// save map coordinates
	if (!load3D && numSpectra>1)
		if (numpnts(w_x)>0)
			Duplicate/O/D w_x $(strbasename+"_mapX") /wave=wMapX
			if (prefs.verbose)
				printf "X coordinates saved as %s_mapX\r", strbasename
			endif
		endif
		if (numpnts(w_y)>0)
			Duplicate/O/D w_y $(strbasename+"_mapY") /wave=wMapY
			if (prefs.verbose)
				printf "Y coordinates saved as %s_mapY\r", strbasename
			endif
		endif
		if (numpnts(w_z)>0)
			Duplicate/O/D w_z $(strbasename+"_mapZ") /wave=wMapZ
			if (prefs.verbose)
				printf "Z coordinates saved as %s_mapZ\r", strbasename
			endif
		endif
	endif
	
	if (WaveExists(img) && haveImageCoords)
		SetScale/I x, ImageCoords[0], ImageCoords[1], img
		SetScale/I y, ImageCoords[2], ImageCoords[3], img
		if (PlotMap && !load3D) // plot image and overlay map points
			if (haveMapCoords)
				makeMapBrowser(strSelect=strBaseName)
			else
				Display/K=1; AppendImage img
				ModifyGraph standoff=0, height={Plan,1,left,bottom}
				SetAxis/A/R left
			endif
		endif
		
		if (load3D && waveexists(w3D))
			strNote = note(w3D)
			strNote = ReplaceStringByKey("ImgWave", strNote, nameofwave(img), ":", "\r")
			note/K w3D, strNote
		endif
		
	endif
	
	if (PlotMap && load3D && waveexists(w3D))
		Execute/Z "ProcGlobal#Hyperspec#MakeHyperspecBrowser(" + GetWavesDataFolder(w3D, 2) + ")"
	endif

	return 1
end

function readImageCoords(ImageCoords, WXDMstart, WXDMend, refnum, prefs, [mapLimits])
	wave ImageCoords
	variable WXDMstart, WXDMend, refnum
	STRUCT PackagePrefs &prefs
	wave mapLimits
	
	Make/B/U/free/N=8 w8bytes, w_match
	w_match = {77,97,112,65,114,101,97,98} // "MapAreab" ASCII
	variable pos = WXDMstart+8
	FSetPos refnum, WXDMstart+8
	string oneByte=" "
	for (pos=WXDMstart+8;pos<=WXDMend;pos+=1)
		FBinRead/B=3 refnum, oneByte
		if (stringmatch(oneByte, "M"))
			FSetPos refnum, pos
			FBinRead/B=3 refnum, w8bytes
			w8bytes -= w_match
			if (sum(w8bytes) == 0)
				break
			endif
			FSetPos refnum, pos+1
		endif
	endfor
	if (pos >= WXDMend)
		return 0
	endif
	variable start = pos
	
	if (prefs.verbose > 1)
		Print/D "MapArea found at byte", start
	endif
	
	if (ParamIsDefault(mapLimits) == 0) // try using map limits to find sensible image coordinates
		variable xmax, xmin, ymax, ymin, numtest, i, logMin, logMax
		Make/free/d/N=4 w_numtest
		logMin = -6 // image coordinates expected to have resolution of about 6 significant figures
		logMax = 6 // image coordinates expected to have magnitude < 1,000,000
		for(pos=start+8;pos<WXDMend-32;pos+=4)
			FSetPos refnum, pos
			FBinRead/B=3/F=5 refnum, ImageCoords
			
			w_numtest = log(abs(ImageCoords))
			// if image has a zero-coordinate this will fail
			if (WaveMin(w_numtest)<(logMin) || WaveMax(w_numtest)>(logMax))
				continue
			endif
			// looks like we have found a sequence of four sensible double-precision numbers
			// check to see that map coordinates fall within the image boundary
			xmax = max(ImageCoords[0],ImageCoords[1])
			xmin = min(ImageCoords[0],ImageCoords[1])
			ymax = max(ImageCoords[2],ImageCoords[3])
			ymin = min(ImageCoords[2],ImageCoords[3])
			if (xmin<mapLimits[0] && xmax>mapLimits[1] && ymin<mapLimits[2] && ymax>mapLimits[3])
				if (prefs.verbose)
					printf "Image coordinates = (%g,%g,%g,%g)\r", ImageCoords[0],ImageCoords[1],ImageCoords[2],ImageCoords[3]
				endif
				return 1
			endif
		endfor
	endif
	// brute force method has failed
		
	// try to use a binary signature to locate coordinates
	// this is pure guesswork
	w_match = {0,0,0,255,255,255,255,1}
	for (pos=start;pos<WXDMend;pos+=1)
		FSetPos refnum, pos
		FBinRead/B=3 refnum, w8bytes
		w8bytes -= w_match
		if (sum(w8bytes) == 0)
			break
		endif
	endfor
	
	if (pos > (WXDMend-32)) // try again with a different binary signature
		w_match = {0,0,0,1,0,0,0,2,0,0,0,0,0,0,0,1}
		Make/B/U/free/N=16 w16bytes
		for(pos=start;pos<WXDMend;pos+=1)
			FSetPos refnum, pos
			FBinRead/B=3 refnum, w16bytes
			w16bytes -= w_match
			if (sum(w16bytes) == 0)
				break
			endif
		endfor
	endif
	
	if (pos > (WXDMend-32)) // give up
		return 0
	endif
	
	if (prefs.verbose > 1)
		Print/D "Image coordinates signature found at byte", pos
	endif

	FBinRead/B=3/F=5 refnum, ImageCoords
	if (prefs.verbose)
		printf "Image coordinates = (%g,%g,%g,%g)\r", ImageCoords[0],ImageCoords[1],ImageCoords[2],ImageCoords[3]
	endif
	return 1
end

function readXYZpositions(w_x, w_y, w_z, numSpectra, blockStart, blockEnd, refnum, prefs)
	wave w_x, w_y, w_z
	variable numSpectra, blockStart, blockEnd, refnum
	STRUCT PackagePrefs &prefs
			
	variable pos
	Make/B/U/free/N=8 w8bytes, w_match
	
	w_match = {5,0,0,0,88,0,0,0} // 88 = X
	for(pos=blockStart+16;Pos<blockEnd-16;pos+=4)
		FSetPos refnum, pos
		FBinRead/B=3 refnum, w8bytes
		w8bytes -= w_match
		if (sum(w8bytes) == 0)
			Redimension/D/N=(numSpectra) w_x
			FSetPos refnum, pos+20
			FBinRead/B=3/F=5 refnum, w_x
			break
		endif
	endfor
	w_match = {5,0,0,0,89,0,0,0} // 89 = Y
	for (pos=blockStart+16;Pos<blockEnd-16;pos+=4)
		FSetPos refnum, pos
		FBinRead/B=3 refnum, w8bytes
		w8bytes -= w_match
		if (sum(w8bytes) == 0)
			Redimension/N=(numSpectra)/D w_y
			FSetPos refnum, pos+20
			FBinRead/B=3/F=5 refnum, w_y
			break
		endif
	endfor
	w_match = {5,0,0,0,90,0,0,0}  // 90 = Z
	for(pos=blockStart+16;Pos<blockEnd-16;pos+=4)
		FSetPos refnum, pos
		FBinRead/B=3 refnum, w8bytes
		w8bytes -= w_match
		if (sum(w8bytes) == 0)
			Redimension/N=(numSpectra)/D w_z
			FSetPos refnum, pos+20
			FBinRead/B=3/F=5 refnum, w_z
			break
		endif
	endfor
	
	if (prefs.verbose > 1)
		printf "Read %d X, %d Y, %d Z coordinates from ORGN block\r", numpnts(w_x), numpnts(w_y), numpnts(w_z)
	endif
	
	return 1
end

// turns raw jpeg binary in w_jpeg into an image wave
function/WAVE SaveImage(wave/B w_jpeg, string strBaseName, STRUCT PackagePrefs &prefs)
	
	variable refnum
	string tempPath = SpecialDirPath("Temporary", 0, 0, 0)
	Open/Z refnum as tempPath+"WiREJPEG.jpg"
	FBinWrite refnum, w_jpeg
	Close refnum
	ImageLoad/Q/Z/T=jpeg/N=$(strBaseName+"_img")/O tempPath+"WiREJPEG.jpg"
	if (V_flag)
		wave img = $StringFromList(0, S_waveNames)
		if (prefs.verbose)
			printf "%d x %d image saved as %s\r", DimSize(img, 0), DimSize(img, 1), NameOfWave(img)
		endif
		return img
	endif
	return $""
end

function/S DoAlertString(string OverwriteList)	
	
	string cmd = "The following waves will be overwritten:\r"
	string tooMany = "\r*** TOO MANY WAVES TO LIST ***\r"
	string cont = "\rContinue?"
	variable cmdLen = strlen(cmd)+strlen(tooMany)+strlen(cont)
	variable i
	variable numWaves = ItemsInList(OverwriteList)
	string ithWave
	for (i=0;i<numWaves;i+=1)
		ithWave = StringFromList(i, OverwriteList)
		// avoid too-long alert
		if ( cmdLen+strlen(ithWave+", ") >= 1023 )
			cmd = RemoveEnding(cmd, ", ")+tooMany
			break
		endif
		cmd += ithWave+", "
		cmdLen += strlen(ithWave+", ")
	endfor
	cmd = RemoveEnding(cmd, ", ")+cont
	return cmd
end

// --------------------- map browser functions -------------------

function makeMapBrowser([string strSelect])
	
	STRUCT PackagePrefs prefs
	LoadPrefs(prefs)
	
	if (prefs.options & 1)
		string cmd = "WiRE loader is set to load maps as 3D waves, which can be plotted using the Hyperspectral Browser package"
		if (!strlen(WinList("hyperspec*", ";", "INDEPENDENTMODULE:1")))
			cmd += " (available for download from IgorExchange)"
		endif
		cmd += ". To use the old WiRE map browser, edit file loading options via Data > Load Waves > Packages > Wire Loader Settings..."
		doalert 0, cmd
		return 0
	endif
	
	KillWindow/Z MapBrowserPanel
	NewDataFolder/O root:Packages
	NewDataFolder/O root:Packages:MapBrowser
	DFREF dfr = root:Packages:MapBrowser
	Make/O/N=1 dfr:pointY /wave=pointY, dfr:pointX /wave=pointX
	pointY = NaN; pointX = NaN
	variable/G dfr:selectedPoint=NaN
	Make/O/N=0/T dfr:toPlot
	Make/O/N=0/WAVE dfr:allWaves
	Make/O dfr:preview /wave=preview, dfr:previewX /wave=previewX
	preview = NaN; previewX = NaN
	
	NewPanel/K=1/N=MapBrowserPanel/W=(50,50,860,390) as "Map Browser"
	
	DefineGuide/W=MapBrowserPanel guidePreviewRight={FR,-20}
	DefineGuide/W=MapBrowserPanel guideMapLeft={FL,20}
	DefineGuide/W=MapBrowserPanel guideCentre={FL,0.5,FR}
	DefineGuide/W=MapBrowserPanel guideMapRight={guideCentre, -10}
	DefineGuide/W=MapBrowserPanel guidePreviewLeft={guideCentre, 10}
	DefineGuide/W=MapBrowserPanel guideTop={FT,20}
	DefineGuide/W=MapBrowserPanel guideMapBottom={FB,-80}
	DefineGuide/W=MapBrowserPanel guidePreviewBottom={guideMapBottom,40}
		
	Display/HOST=MapBrowserPanel/FG=(guideMapLeft,guideTop,guideMapRight,guideMapBottom)/PG=(guideMapLeft,guideTop,guideMapRight,guideMapBottom)  /N=MapPlot
	AppendToGraph/W=MapBrowserPanel#MapPlot pointY vs pointX
	ModifyGraph/W=MapBrowserPanel#MapPlot marker(pointY)=1,rgb(pointY)=(0,65535,0)
	
	Display/HOST=MapBrowserPanel/FG=(guidePreviewLeft,guideTop,guidePreviewRight,guidePreviewBottom)/PG=(guidePreviewLeft,guideTop,guidePreviewRight,guideMapBottom) /N=PreviewPlot
	AppendToGraph/W=MapBrowserPanel#PreviewPlot preview vs previewX
	ModifyGraph/W=MapBrowserPanel#PreviewPlot rgb(preview)=(65535,0,0)
	ModifyGraph/W=MapBrowserPanel#PreviewPlot frameStyle=0, noLabel(left)=2, wbRGB=(61166,61166,61166)

	variable vMode = 1
	if (ParamIsDefault(strSelect) == 0)
		vMode = 1 + WhichListItem(strSelect+"_img", WaveList("*_img",";","DIMS:3"))
		vMode = (vMode == 0) ? 1 : vMode
	endif
	PopupMenu popupImage,win=MapBrowserPanel,pos={20,270},size={124,23}, mode=vMode, title="Map", fsize=14
	PopupMenu popupImage,win=MapBrowserPanel,value=WaveList("*_img",";","DIMS:3"), Proc=MapBrowserPopupProc
	Button buttonSelectAll,win=MapBrowserPanel,pos={20,300},size={70,25},title="Select All",Proc=MapBrowserButtonProc
	Button buttonClear,win=MapBrowserPanel,pos={110,300},size={45,25},title="Clear",Proc=MapBrowserButtonProc
	Button buttonNewGraph,win=MapBrowserPanel,pos={792-80,300},size={80,25},title="New Graph"
	Button buttonNewGraph,win=MapBrowserPanel,disable=2,Proc=MapBrowserButtonProc
	Button buttonAppendToTop,win=MapBrowserPanel,pos={792-80-20-100,300},size={100,25},title="Append to Top"
	Button buttonAppendToTop,win=MapBrowserPanel,disable=2, Proc=MapBrowserButtonProc
	
	SetWindow MapBrowserPanel, activeChildFrame=0
	SetWindow MapBrowserPanel, hook(MBhook)=WireLoader#MapBrowserHook
	SetActiveSubwindow MapBrowserPanel
	
	// resizing user data for controls
	PopupMenu popupImage,win=MapBrowserPanel,userdata(ResizeControlsInfo)= A"!!,BY!!#BA!!#>Z!!#<Xz!!#](Aon\"Qzzzzzzzzzzzzzz!!#](Aon\"Qzz"
	PopupMenu popupImage,win=MapBrowserPanel,userdata(ResizeControlsInfo) += A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
	PopupMenu popupImage,win=MapBrowserPanel,userdata(ResizeControlsInfo) += A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
	Button buttonSelectAll,win=MapBrowserPanel,userdata(ResizeControlsInfo)= A"!!,BY!!#BP!!#?E!!#=+z!!#](Aon\"Qzzzzzzzzzzzzzz!!#](Aon\"Qzz"
	Button buttonSelectAll,win=MapBrowserPanel,userdata(ResizeControlsInfo) += A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
	Button buttonSelectAll,win=MapBrowserPanel,userdata(ResizeControlsInfo) += A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
	Button buttonClear,userdata(ResizeControlsInfo)= A"!!,FA!!#BP!!#>B!!#=+z!!#](Aon\"Qzzzzzzzzzzzzzz!!#](Aon\"Qzz"
	Button buttonClear,userdata(ResizeControlsInfo) += A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
	Button buttonClear,userdata(ResizeControlsInfo) += A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
	Button buttonNewGraph,win=MapBrowserPanel,userdata(ResizeControlsInfo)= A"!!,JC!!#BP!!#?Y!!#=+z!!#o2B4uAezzzzzzzzzzzzzz!!#o2B4uAezz"
	Button buttonNewGraph,win=MapBrowserPanel,userdata(ResizeControlsInfo) += A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
	Button buttonNewGraph,win=MapBrowserPanel,userdata(ResizeControlsInfo) += A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
	Button buttonAppendToTop,win=MapBrowserPanel,userdata(ResizeControlsInfo)= A"!!,J%!!#BP!!#@,!!#=+z!!#o2B4uAezzzzzzzzzzzzzz!!#o2B4uAezz"
	Button buttonAppendToTop,win=MapBrowserPanel,userdata(ResizeControlsInfo) += A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
	Button buttonAppendToTop,win=MapBrowserPanel,userdata(ResizeControlsInfo) += A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
	
	// resizing user data for panel window
	SetWindow MapBrowserPanel,userdata(ResizeControlsInfo)= A"!!*'\"z!!#DZJ,hs:zzzzzzzzzzzzzzzzzzzzz"
	SetWindow MapBrowserPanel,userdata(ResizeControlsInfo) += A"zzzzzzzzzzzzzzzzzzzzzzzzz"
	SetWindow MapBrowserPanel,userdata(ResizeControlsInfo) += A"zzzzzzzzzzzzzzzzzzz!!!"
	
	SetWindow MapBrowserPanel,userdata(ResizeControlsGuides)=  "guidePreviewRight;guideMapLeft;guideCentre;guideMapRight;guidePreviewLeft;guideTop;guideMapBottom;guidePreviewBottom;"
	SetWindow MapBrowserPanel,userdata(ResizeControlsInfoguidePreviewRight)=  "NAME:guidePreviewRight;WIN:MapBrowserPanel;TYPE:User;HORIZONTAL:0;POSITION:790.00;GUIDE1:FR;GUIDE2:;RELPOSITION:-20;"
	SetWindow MapBrowserPanel,userdata(ResizeControlsInfoguideMapLeft)=  "NAME:guideMapLeft;WIN:MapBrowserPanel;TYPE:User;HORIZONTAL:0;POSITION:20.00;GUIDE1:FL;GUIDE2:;RELPOSITION:20;"
	SetWindow MapBrowserPanel,userdata(ResizeControlsInfoguideCentre)=  "NAME:guideCentre;WIN:MapBrowserPanel;TYPE:User;HORIZONTAL:0;POSITION:405.00;GUIDE1:FL;GUIDE2:FR;RELPOSITION:0.5;"
	SetWindow MapBrowserPanel,userdata(ResizeControlsInfoguideMapRight)=  "NAME:guideMapRight;WIN:MapBrowserPanel;TYPE:User;HORIZONTAL:0;POSITION:395.00;GUIDE1:guideCentre;GUIDE2:;RELPOSITION:-10;"
	SetWindow MapBrowserPanel,userdata(ResizeControlsInfoguidePreviewLeft)=  "NAME:guidePreviewLeft;WIN:MapBrowserPanel;TYPE:User;HORIZONTAL:0;POSITION:415.00;GUIDE1:guideCentre;GUIDE2:;RELPOSITION:10;"
	SetWindow MapBrowserPanel,userdata(ResizeControlsInfoguideTop)=  "NAME:guideTop;WIN:MapBrowserPanel;TYPE:User;HORIZONTAL:1;POSITION:20.00;GUIDE1:FT;GUIDE2:;RELPOSITION:20;"
	SetWindow MapBrowserPanel,userdata(ResizeControlsInfoguideMapBottom)=  "NAME:guideMapBottom;WIN:MapBrowserPanel;TYPE:User;HORIZONTAL:1;POSITION:260.00;GUIDE1:FB;GUIDE2:;RELPOSITION:-80;"
	SetWindow MapBrowserPanel,userdata(ResizeControlsInfoguidePreviewBottom)=  "NAME:guidePreviewBottom;WIN:MapBrowserPanel;TYPE:User;HORIZONTAL:1;POSITION:300.00;GUIDE1:guideMapBottom;GUIDE2:;RELPOSITION:40;"
		
	// resizing panel hook
	SetWindow MapBrowserPanel hook(ResizeControls)=ResizeControls#ResizeControlsHook

	// use popup proc to select an image
	ControlInfo/W=MapBrowserPanel popupImage
	STRUCT WMPopupAction s
	s.eventCode = 2
	s.popStr = S_Value
	MapBrowserPopupProc(s)
end

function MapBrowserPopupProc(STRUCT WMPopupAction &s)
	
	if (s.eventCode != 2)
		return 0
	elseif (strlen(s.popStr)==0 || stringmatch(s.popStr, "_none_"))
		return 0
	endif
	
	string basename = RemoveEnding(s.popStr, "img")
	string oldImg = StringFromList(0, ImageNameList("MapBrowserPanel#MapPlot", ";"))
	
	wave/Z NewImg = $s.popStr
	wave/Z NewMap = $basename+"mapY"
	wave/Z NewMapX = $basename+"mapX"
	wave/Z data = $basename+"0"
	wave/Z dataX = $basename+"x"
	
	variable differentMap = (stringmatch(oldImg, s.popStr) == 0)
	
	do
		RemoveImage/W=MapBrowserPanel#MapPlot/Z  $StringFromList(0, ImageNameList("MapBrowserPanel#MapPlot", ";"))
	while(strlen(ImageNameList("MapBrowserPanel#MapPlot", ";")))
	do
		RemoveFromGraph/W=MapBrowserPanel#MapPlot/Z  $StringFromList(0, TraceNameList("MapBrowserPanel#MapPlot", ";", 1))
	while(strlen(TraceNameList("MapBrowserPanel#MapPlot", ";", 1)))

	AppendImage/W=MapBrowserPanel#MapPlot NewImg
	AppendToGraph/W=MapBrowserPanel#MapPlot NewMap /TN=map vs NewMapX
	ModifyGraph/W=MapBrowserPanel#MapPlot rgb=(65535,0,0), highTrip=100000
	SetAxis/A/R/W=MapBrowserPanel#MapPlot left
	SetAxis/A/W=MapBrowserPanel#MapPlot bottom
		
	DFREF dfr = root:Packages:MapBrowser
	
	wave/Z pointY = dfr:pointY
	wave/Z pointX = dfr:pointX
	wave/Z/T toPlot = dfr:toPlot

	AppendToGraph/W=MapBrowserPanel#MapPlot pointY vs pointX
	ModifyGraph/W=MapBrowserPanel#MapPlot mode=3,marker=0
	ModifyGraph/W=MapBrowserPanel#MapPlot rgb(pointY)=(0,65535,0)
		
	Duplicate/O data dfr:preview /wave=preview
	Duplicate/O dataX dfr:previewX /wave=previewX
	preview = NaN
	SetAxis/A/W=MapBrowserPanel#PreviewPlot
	ModifyGraph/W=MapBrowserPanel#PreviewPlot noLabel(left)=2
	
	wave/wave allwaves=dfr:allWaves
	Redimension/N=(numpnts(NewMap)) allWaves
	allWaves=$(GetDataFolder(1)+basename+num2str(p))
	
	if (differentMap)
		do // clear preview plot
			RemoveFromGraph/W=MapBrowserPanel#PreviewPlot/Z  $StringFromList(0, TraceNameList("MapBrowserPanel#PreviewPlot", ";", 1))
		while(strlen(TraceNameList("MapBrowserPanel#PreviewPlot", ";", 1)))
		AppendToGraph/W=MapBrowserPanel#PreviewPlot/C=(65535,0,0) preview vs previewX
		// clear seleced points
		Redimension/N=1 pointY, pointX
		Redimension/N=0 toPlot
		pointX=NaN; pointY=NaN
	endif
	
	return 0
end

function MapBrowserButtonProc(STRUCT WMButtonAction &s)

	variable hist = 1 // choose whether to record graph building commands to history
	
	if (s.eventCode != 2)
		return 0
	endif
		
	strswitch (s.ctrlName)
		case "buttonSelectAll":
			SelectAll(1)
			return 0
		case "buttonClear":
			SelectAll(0)
			return 0
		case "buttonNewGraph":
			Display
			if (hist)
				Print "•Display"
			endif
		case "buttonAppendToTop":
			// make sure we have a top graph
			string strTopGraph = WinName(0, 1)
			if (strlen(strTopGraph) == 0)
				return 0
			endif
			DoWindow/F $strTopGraph
			
			DFREF packageDfr = root:Packages:MapBrowser
			wave/T toPlot = packageDfr:toPlot
			
			variable i, imax = numpnts(toPlot)
			string cmd = ""
			
			for(i=0;i<imax;i+=1)
				wave w = $toPlot[i]
				wave w_x = MapBrowserXWaveRef(w)
				AppendToGraph w vs w_x
				if (hist)
					printf "•AppendToGraph %s vs %s\r", NameOfWave(w), NameOfWave(w_x)
				endif
			endfor
			
			if (stringmatch(s.ctrlName, "buttonNewGraph"))
				Label/Z left "Intensity"
				Label/Z bottom "Wavenumber (cm\\S-1\\M)"
			endif
			break
	endswitch
	return 0
end

function MapBrowserHook(STRUCT WMWinHookStruct &s)
	variable hookResult = 0
	DFREF packageDfr = root:packages:MapBrowser
	NVAR selectedPoint = packageDfr:selectedPoint
		
	switch (s.eventCode)
		case 4: // mousemoved
			wave/Z/SDFR=packageDfr pointY,pointX,preview
			if (numpnts(pointY) == 0)
				break
			endif
			variable pnt = numpnts(pointY)-1
			string keyList = TraceFromPixel(s.mouseloc.h,s.mouseloc.v, "WINDOW:MapBrowserPanel#MapPlot;ONLY:map;")
			if (strlen(keyList))
				wave/Z mapY = TraceNameToWaveRef("MapBrowserPanel#MapPlot", "map")
				wave/Z mapX = XWaveRefFromTrace("MapBrowserPanel#MapPlot", "map")
				DFREF dfr = GetWavesDataFolderDFR(TraceNameToWaveRef("MapBrowserPanel#MapPlot", "map"))
				string basename = RemoveEnding(NameOfWave(mapY), "MapY")
				string strHitP = StringByKey("HITPOINT", keyList)
				selectedPoint = str2num(strHitP)
				wave/Z/SDFR=dfr xwave=$basename+"x"
				wave/Z/SDFR=dfr datawave=$basename+strHitP
							
				if (WaveExists(xwave)==0 || WaveExists(datawave)==0)
					break
				endif
				
				pointY[pnt] = mapY[selectedPoint]
				pointX[pnt] = mapX[selectedPoint]
				preview = datawave
				ReorderTraces/W=MapBrowserPanel#PreviewPlot _front_, {preview}
				ModifyGraph/W=MapBrowserPanel#PreviewPlot noLabel(left)=2
			else // not on point
				pointY[pnt] = NaN; pointX[pnt] = NaN
				selectedPoint = NaN
				preview = NaN
			endif
			break
		case 5: // mouseup
			if (numtype(selectedPoint) == 2) // nan
				return hookResult
			endif
			selectPoint(selectedPoint, -1)
			break
		case 2: // kill
			KillWindow/Z MapBrowserPanel#MapPlot
			KillWindow/Z MapBrowserPanel#PreviewPlot
			KillDataFolder/Z packageDfr
			break
	endswitch
	
	return hookResult	// If non-zero, we handled event and Igor will ignore it.
end

// select/deselect a point from the map by number, add selected spectrum to preview plot
// select = 1 to select, 0 to deselect, -1 to toggle selection
function SelectPoint(variable pnt, variable select)	
	DFREF dfr = root:packages:MapBrowser
	wave/Z/SDFR=dfr pointX, pointY
	wave/Z/T/SDFR=dfr toPlot
	wave/Z/Wave/SDFR=dfr allWaves
	wave/Z mapY = TraceNameToWaveRef("MapBrowserPanel#MapPlot", "map")
	wave/Z mapX = XWaveRefFromTrace("MapBrowserPanel#MapPlot", "map")
	string strName = NameOfWave(allWaves[pnt])
		
	FindValue/TEXT=strName/TXOP=4 toPlot
	if (select == -1) // toggle selection
		select = !(V_Value > -1)
	endif
	if (select == (V_Value > -1)) // already selected/deselected
		return 0
	endif
	
	if (select == 0)
		DeletePoints V_Value, 1, toPlot, pointX, pointY
		RemoveFromGraph/W=MapBrowserPanel#PreviewPlot $strName
		wave preview = dfr:preview
		preview = NaN
	else
		wave w = allWaves[pnt]
		toPlot[numpnts(toPlot)] = {NameOfWave(w)}
		pointX[numpnts(pointX)-1] = mapX[pnt]
		pointY[numpnts(pointY)-1] = mapY[pnt]
		pointX[numpnts(pointX)] = {NaN}; pointY[numpnts(pointY)] = {NaN}
		AppendToGraph/C=(2,39321,1)/W=MapBrowserPanel#PreviewPlot w vs MapBrowserXWaveRef(w)
		ModifyGraph/W=MapBrowserPanel#PreviewPlot noLabel(left)=2
	endif
	
	Button buttonNewGraph,win=MapBrowserPanel, disable=2*(numpnts(toPlot)==0)
	Button buttonAppendToTop,win=MapBrowserPanel, disable=2*(numpnts(toPlot)==0)
	return 1
end

// returns waveref for the X wave paired with wave w
function/WAVE MapBrowserXWaveRef(wave w)
	if (WaveExists(w) == 0)
		return $""
	endif
	DFREF dfr = GetWavesDataFolderDFR(w)
	string s_x = StringByKey("Xwave", note(w), ":", "\r")
	if (cmpstr(s_x[0]," ") == 0)
		do // remove whitespace after key separator
			s_x = ReplaceString(" ", s_x, "", 0, 1)
		while ((cmpstr(s_x[0]," ")==0) && strlen(s_x))
	endif
	do
		s_x = RemoveEnding(s_x, " ")
	while (cmpstr(s_x[strlen(s_x)-1]," ") == 0)
	
	return dfr:$s_x
end

menu "GraphMarquee", dynamic	
	MenuString(1), /Q, SelectPointsInMarquee(1)
	MenuString(0), /Q, SelectPointsInMarquee(0)
end

function/S MenuString(variable select)
	if (cmpstr(WinList("*", ";", "WIN:"), "MapBrowserPanel;") == 0)
		return SelectString(select, "Deselect", "Select") + " points in marquee"
	endif
	return ""
end

function SelectPointsInMarquee(variable select)	
	wave/Z mapY = TraceNameToWaveRef("MapBrowserPanel#MapPlot", "map")
	wave/Z mapX = XWaveRefFromTrace("MapBrowserPanel#MapPlot", "map")
	Make/free/N=0 xPoly, yPoly
	GetMarquee/K/Z left, bottom
	xPoly = {v_left, v_left, v_right, v_right}
	yPoly = {v_top, v_bottom, v_bottom, v_top}
	FindPointsInPoly mapX, mapY, xPoly, yPoly
	wave W_inPoly = W_inPoly
	W_inPoly = W_inPoly ? selectPoint(p, select) : 0
end

function SelectAll(variable select)	
	wave/Z mapY = TraceNameToWaveRef("MapBrowserPanel#MapPlot", "map")
	Make/free/N=(numpnts(mapY)) w = selectPoint(p, select) // loop over points of map
end

// ProcedureVersion("") doesn't work for independent modules!
// extract procedure version from this file
static function GetThisVersion()

	variable procVersion
	// try the quick way
	#if exists("ProcedureVersion") == 3
	procVersion = ProcedureVersion("")
	if (procVersion)
		return round(100 * procVersion)
	endif
	#endif
	
	int maxLines = 20 // number of lines to search for version pragma
	int refNum, i
	string strHeader = ""
	string strLine = ""
	
	Open/R/Z refnum as FunctionPath("")
	if (refnum == 0)
		return 0
	endif
	for (i=0;i<maxLines;i+=1)
		FReadLine refNum, strLine
		strHeader += strLine
	endfor
	Close refnum
	wave/T ProcText = ListToTextWave(strHeader, "\r")
	Grep/Q/E="(?i)^#pragma[\s]*version[\s]*=" /LIST/Z ProcText
	if (v_flag != 0)
		return 0
	endif
	s_value = LowerStr(TrimString(s_value, 1))
	sscanf s_value, "#pragma version = %f", procVersion
	ProcVersion = (V_flag!=1 || ProcVersion<=0) ? 0 : ProcVersion

	return round(100 * ProcVersion)
end

function GetScreenOrigin(STRUCT point &pnt)
	
	// set default value
	pnt.h = 0
	pnt.v = 0
	
	GetMouse
	
	#ifdef debug
	Print v_left, v_top
	#endif

	STRUCT point mousePoint
	mousePoint.h = v_left
	mousePoint.v = v_top
	
	STRUCT rect screenRect
	
	string strInfo = IgorInfo(0)
	string str = ""
	int numScreens = NumberByKey("NSCREENS", strInfo)
	int i, left, top, right, bottom
	
	for (i=1;i<=numScreens;i++)
		str = StringByKey("SCREEN"+num2str(i), strInfo)
		sscanf str, "DEPTH=%*d,RECT=%d,%d,%d,%d", left, top, right, bottom
		if (V_flag != 4)
			continue
		endif
		screenRect.left   = left
		screenRect.top    = top
		screenRect.right  = right
		screenRect.bottom = bottom
		if (PointInRect(mousePoint, screenRect))
			pnt.h = screenRect.left
			pnt.v = screenRect.top
			return 1
		endif
	endfor
	return 0
end

// point and rect structures must have same units
function PointInRect(STRUCT point &pnt, STRUCT rect &r)
	return ( pnt.h>r.left && pnt.h<r.right && pnt.v>r.top && pnt.v<r.bottom )
end